
#ifndef FACTOR_GNSS_FIX_2D_H_
#define FACTOR_GNSS_FIX_2D_H_

//Wolf includes
#include "core/common/wolf.h"
#include "core/factor/factor_autodiff.h"
#include "core/frame/frame_base.h"
#include "gnss/sensor/sensor_gnss.h"

namespace wolf {
    
WOLF_PTR_TYPEDEFS(FactorGnssFix2D);

class FactorGnssFix2D : public FactorAutodiff<FactorGnssFix2D, 3, 2, 1, 3, 3, 1, 1, 1>
{
    protected:
        SensorGnssPtr sensor_gnss_ptr_;

    public:

        FactorGnssFix2D(FeatureBasePtr& _ftr_ptr, const SensorGnssPtr& _sensor_gnss_ptr, const ProcessorBasePtr& _processor_ptr, bool _apply_loss_function = false, FactorStatus _status = FAC_ACTIVE) :
            FactorAutodiff<FactorGnssFix2D, 3, 2, 1, 3, 3, 1, 1, 1>("GNSS FIX 2D",
                                                                     nullptr,
                                                                     nullptr,
                                                                     nullptr,
                                                                     nullptr,
                                                                     _processor_ptr,
                                                                     _apply_loss_function,
                                                                     _status,
                                                                     _ftr_ptr->getFrame()->getP(),
                                                                     _ftr_ptr->getFrame()->getO(),
                                                                     _sensor_gnss_ptr->getStateBlock(0),
                                                                     _sensor_gnss_ptr->getEnuMapTranslationState(),
                                                                     _sensor_gnss_ptr->getEnuMapRollState(),
                                                                     _sensor_gnss_ptr->getEnuMapPitchState(),
                                                                     _sensor_gnss_ptr->getEnuMapYawState()),
            sensor_gnss_ptr_(_sensor_gnss_ptr)
        {
            WOLF_WARN_COND(!sensor_gnss_ptr_->isEnuDefined() && _status == FAC_ACTIVE, "Creating an active GNSS Fix 2D factor without initializing ENU");
        }

        virtual ~FactorGnssFix2D() = default;

        template<typename T>
        bool operator ()(const T* const _x,
                         const T* const _o,
                         const T* const _x_antena,
                         const T* const _t_ENU_map,
                         const T* const _roll_ENU_map,
                         const T* const _pitch_ENU_map,
                         const T* const _yaw_ENU_map,
                         T* _residuals) const;

};

template<typename T>
inline bool FactorGnssFix2D::operator ()(const T* const _x,
                                         const T* const _o,
                                         const T* const _x_antena,
                                         const T* const _t_ENU_map,
                                         const T* const _roll_ENU_map,
                                         const T* const _pitch_ENU_map,
                                         const T* const _yaw_ENU_map,
                                         T* _residuals) const
{
    Eigen::Map<const Eigen::Matrix<T,2,1> > t_map_base(_x);
    Eigen::Matrix<T,2,2> R_map_base = Eigen::Rotation2D<T>(_o[0]).matrix();
    Eigen::Map<const Eigen::Matrix<T,3,1> > t_base_antena(_x_antena);
    Eigen::Map<Eigen::Matrix<T,3,1> > residuals_ECEF(_residuals);

    Eigen::Matrix<T,3,3> R_ENU_ECEF = sensor_gnss_ptr_->getREnuEcef().cast<T>();
    Eigen::Matrix<T,3,1> t_ENU_ECEF = sensor_gnss_ptr_->gettEnuEcef().cast<T>();
    Eigen::Matrix<T,2,3> R_map_ENU = sensor_gnss_ptr_->computeREnuMap(_roll_ENU_map[0], _pitch_ENU_map[0], _yaw_ENU_map[0]).transpose().topRows(2);
    Eigen::Map<const Eigen::Matrix<T,3,1> > t_ENU_map(_t_ENU_map);
    Eigen::Matrix<T,3,1> fix_ECEF = getMeasurement().cast<T>();

    //std::cout << "computeREnuMap():\n" << sensor_gnss_ptr_->computeREnuMap((const double) _roll[0], (const double) _pitch[0], (const double) _yaw[0]).transpose().topRows(2) << std::endl;
    //std::cout << "getREnuEcef():\n" << sensor_gnss_ptr_->gettEnuEcef() << std::endl;
    //std::cout << "gettEnuEcef(): " << sensor_gnss_ptr_->gettEnuEcef().transpose() << std::endl;

    // Transform ECEF 3D Fix Feature to 2D Fix in Map coordinates (removing z)
    //std::cout << "R_map_ENU:\n\t" << R_map_ENU(0,0) << "\t" << R_map_ENU(0,1) << "\t" << R_map_ENU(0,2) << "\n\t" << R_map_ENU(1,0) << "\t" << R_map_ENU(1,1) << "\t" << R_map_ENU(1,2) << std::endl;
    //std::cout << "R_ENU_ECEF:\n\t" << R_ENU_ECEF(0,0) << "\t" << R_ENU_ECEF(0,1) << "\t" << R_ENU_ECEF(0,2) << "\n\t" << R_ENU_ECEF(1,0) << "\t" << R_ENU_ECEF(1,1) << "\t" << R_ENU_ECEF(1,2) << "\n\t" << R_ENU_ECEF(2,0) << "\t" << R_ENU_ECEF(2,1) << "\t" << R_ENU_ECEF(2,2) << std::endl;
    //std::cout << "t_ENU_ECEF:\n\t" << t_ENU_ECEF(0) << "\n\t" << t_ENU_ECEF(1) << "\n\t" << t_ENU_ECEF(2) << std::endl;
    Eigen::Matrix<T,2,1> fix_map = R_map_ENU * (R_ENU_ECEF * fix_ECEF + t_ENU_ECEF - t_ENU_map);
    //std::cout << "fix 3D:\n\t" << getMeasurement()(0) << "\n\t" << getMeasurement()(1) << "\n\t" << getMeasurement()(2) << std::endl;
    //std::cout << "fix_map:\n\t" << fix_map[0] << "\n\t" << fix_map[1] << std::endl;

    // Antena position in Map coordinates
    Eigen::Matrix<T,2,1> antena_map = R_map_base * t_base_antena.head(2) + t_map_base;
    //std::cout << "antena_map:\n\t" << antena_map[0] << "\n\t" << antena_map[1] << std::endl;

    // Compute residual rotating information matrix to 2D Map coordinates
    // Covariance & information are rotated by R*S*R' so for the squred root upper (or right) R*L*U*R'->U*R'
    // In this case R = R_map_ENU * R_ENU_ECEF, then R' = R_ENU_ECEF' * R_map_ENU'
    residuals_ECEF = (getMeasurementSquareRootInformationUpper().cast<T>() * R_ENU_ECEF.transpose() * R_map_ENU.transpose()) * (antena_map - fix_map);
    //std::cout << "residuals_ECEF:\n\t" << _residuals[0] << "\n\t" << _residuals[1] << "\n\t" << _residuals[2]<< std::endl;

    return true;
}

} // namespace wolf

#endif
