
#ifndef FACTOR_GNSS_FIX_3D_H_
#define FACTOR_GNSS_FIX_3D_H_

//Wolf includes
#include "core/common/wolf.h"
#include "core/factor/factor_autodiff.h"
#include "core/frame/frame_base.h"

namespace wolf {
    
WOLF_PTR_TYPEDEFS(FactorGnssFix3D);

class FactorGnssFix3D : public FactorAutodiff<FactorGnssFix3D, 3, 3, 4, 3, 3, 1, 1, 1>
{
    protected:
        SensorGnssPtr sensor_gnss_ptr_;

    public:

        FactorGnssFix3D(FeatureBasePtr& _ftr_ptr, const SensorGnssPtr& _sensor_gnss_ptr, const ProcessorBasePtr& _processor_ptr, bool _apply_loss_function = false, FactorStatus _status = FAC_ACTIVE) :
            FactorAutodiff<FactorGnssFix3D, 3, 3, 4, 3, 3, 1, 1, 1>("GNSS FIX 3D",
                                                                     nullptr,
                                                                     nullptr,
                                                                     nullptr,
                                                                     nullptr,
                                                                     _processor_ptr,
                                                                     _apply_loss_function,
                                                                     _status,
                                                                     _ftr_ptr->getFrame()->getP(),
                                                                     _ftr_ptr->getFrame()->getO(),
                                                                     _sensor_gnss_ptr->getStateBlock(0),
                                                                     _sensor_gnss_ptr->getEnuMapTranslationState(),
                                                                     _sensor_gnss_ptr->getEnuMapRollState(),
                                                                     _sensor_gnss_ptr->getEnuMapPitchState(),
                                                                     _sensor_gnss_ptr->getEnuMapYawState()),
            sensor_gnss_ptr_(_sensor_gnss_ptr)
        {
            WOLF_WARN_COND(!sensor_gnss_ptr_->isEnuDefined() && _status == FAC_ACTIVE, "Creating an ACTIVE GNSS Fix 3D factor without initializing ENU");
        }

        virtual ~FactorGnssFix3D() = default;

        template<typename T>
        bool operator ()(const T* const _x,
                         const T* const _o,
                         const T* const _x_antena,
                         const T* const _t_ENU_map,
                         const T* const _roll_ENU_map,
                         const T* const _pitch_ENU_map,
                         const T* const _yaw_ENU_map,
                         T* _residuals) const;

};

template<typename T>
inline bool FactorGnssFix3D::operator ()(const T* const _x,
                                         const T* const _o,
                                         const T* const _x_antena,
                                         const T* const _t_ENU_map,
                                         const T* const _roll_ENU_map,
                                         const T* const _pitch_ENU_map,
                                         const T* const _yaw_ENU_map,
                                         T* _residuals) const
{
    Eigen::Map<const Eigen::Matrix<T,3,1> > t_map_base(_x);
    Eigen::Map<const Eigen::Quaternion<T> > q_map_base(_o);
    Eigen::Map<const Eigen::Matrix<T,3,1> > t_base_antena(_x_antena);
    Eigen::Map<Eigen::Matrix<T,3,1> > residuals_ECEF(_residuals);

    Eigen::Matrix<T,3,3> R_ENU_ECEF = sensor_gnss_ptr_->getREnuEcef().cast<T>();
    Eigen::Matrix<T,3,1> t_ENU_ECEF = sensor_gnss_ptr_->gettEnuEcef().cast<T>();
    Eigen::Matrix<T,3,3> R_map_ENU = sensor_gnss_ptr_->computeREnuMap(_roll_ENU_map[0], _pitch_ENU_map[0], _yaw_ENU_map[0]).transpose();
    Eigen::Map<const Eigen::Matrix<T,3,1> > t_ENU_map(_t_ENU_map);
    Eigen::Matrix<T,3,1> fix_ECEF = getMeasurement().cast<T>();

    // Transform ECEF 3D Fix Feature to Map coordinates
    Eigen::Matrix<T,3,1> fix_map = R_map_ENU * (R_ENU_ECEF * fix_ECEF + t_ENU_ECEF - t_ENU_map);

    // Antena position in Map coordinates
    Eigen::Matrix<T,3,1> antena_map = q_map_base * t_base_antena + t_map_base;

    // Compute residual rotating information matrix in Map coordinates
    // Covariance & information are rotated by R*S*R' so for the squred root upper (or right) R*L*U*R'->U*R'
    // In this case R = R_map_ENU * R_ENU_ECEF, then R' = R_ENU_ECEF' * R_map_ENU'
    residuals_ECEF = (getMeasurementSquareRootInformationUpper().cast<T>() * R_ENU_ECEF.transpose() * R_map_ENU.transpose()) * (antena_map - fix_map);

    return true;
}

} // namespace wolf

#endif
